import inspect
from time import sleep
from typing import Dict

from appium.webdriver.common.appiumby import AppiumBy

from puma.apps.android import log_action
from puma.apps.android.appium_actions import supported_version, AndroidAppiumActions
from puma.apps.android.teleguard import logger

APPLICATION_PACKAGE = 'ch.swisscows.messenger.teleguardapp'

@supported_version("4.0.7")
class TeleguardActions(AndroidAppiumActions):
    def __init__(self,
                 device_udid,
                 desired_capabilities: Dict[str, str] = None,
                 implicit_wait=1,
                 appium_server='http://localhost:4723'):
        """
        Class with an API for TeleGuard using Appium. Can be used with an emulator or real device attached to the computer.
        """
        AndroidAppiumActions.__init__(self,
                                      device_udid,
                                      APPLICATION_PACKAGE,
                                      desired_capabilities=desired_capabilities,
                                      implicit_wait=implicit_wait,
                                      appium_server=appium_server)
        self.package_name = APPLICATION_PACKAGE


    def _if_chat_go_to_chat(self, chat: str):
        """
        Go to chat if supplied.
        :param chat: Name of the chat to go to, this is either the contact name or the group name.
        """
        if chat is None:
            logger.warning("No chat was supplied. Assuming you are in the correct conversation screen now.")
        else:
            self.select_chat(chat)
            sleep(1)
            if not self._currently_in_conversation(chat):
                raise Exception('Expected to be in conversation screen now, but screen contents are unknown')

    def _currently_at_homescreen(self) -> bool:
        """
        Check whether currently at homescreen of the application.
        :return: boolean if at homescreen
        """
        return (self.is_present('//android.view.View[@content-desc="TeleGuard"]') and
                self.is_present('//android.view.View[@content-desc="Online"]'))

    def _currently_in_conversation(self, chat: str) -> bool:
        """
        Check if currently in a given conversation.
        :return: boolean if in conversation screen or not
        """
        # TeleGuard doesn't contain very descriptive elements, so looking explicitly at the subject of the chat is the
        # only way to identify if you are in the conversation screen.
        return self.is_present(f'//android.view.View[contains(lower-case(@content-desc), "{chat.lower()}")]')

    def return_to_homescreen(self, attempts: int = 10):
        """
        Returns to the start screen of Telegram.
        :param attempts: Number of attempts to return to home screen. Avoids an infinite loop when a popup occurs.
        """
        while not self._currently_at_homescreen() and attempts > 0:
            if self.driver.current_package != self.app_package:
                self.activate_app()
            self.driver.back()
            attempts -= 1
        if not self._currently_at_homescreen():
            raise Exception('Tried to return to homescreen but ran out of attempts...')
        sleep(0.5)

    @log_action
    def select_chat(self, chat: str):
        """
        Opens a given conversation based on the (partial) name of a chat.
        Matching is case-insensitive.
        :param chat: (part of) the conversation name to open
        """
        self.return_to_homescreen()
        xpath = f'//android.widget.ImageView[contains(lower-case(@content-desc), "{chat.lower()}")] | \
        //android.view.View[contains(lower-case(@content-desc), "{chat.lower()}")]'

        self.driver.find_element(by=AppiumBy.XPATH, value=xpath).click()

    @log_action
    def send_message(self, message: str, chat: str = None):
        """
        Send a message in the current or given chat.
        It is recommended to not use the chat param when already in the conversation, as using it will always cause
        puma to navigate out and back into the conversation.
        :param message: The text message to send
        :param chat: Optional: The chat conversation in which to send this message, if not currently in the desired chat
        """
        self._if_chat_go_to_chat(chat)
        text_box_el = self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.EditText[contains(lower-case(@hint), "send a message")]')
        text_box_el.click()
        text_box_el.send_keys(message)
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.FrameLayout[@resource-id="android:id/content"]/android.widget.FrameLayout/android.widget.FrameLayout/android.view.View/android.view.View/android.view.View/android.view.View/android.widget.ImageView[3]').click()

    @log_action
    def add_contact(self, id: str):
        """
        Add a contact by TeleGuard ID.
        :param id: The teleguard ID
        """
        self.return_to_homescreen()
        hamburger = self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.FrameLayout[@resource-id="android:id/content"]/android.widget.FrameLayout/android.widget.FrameLayout/android.view.View/android.view.View/android.view.View/android.view.View/android.view.View[1]/android.view.View[2]/android.view.View[3]')
        hamburger.click()
        add_contact_btn = self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.ImageView[@content-desc="Add contact"]')
        add_contact_btn.click()
        text_box_el = self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.EditText')
        text_box_el.send_keys(id)
        invite_btn = self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.Button[@content-desc="INVITE"]')
        invite_btn.click()

    @log_action
    def accept_invite(self):
        """
        Accept an invite from another user. If you have multiple invites, only one invite will be accepted (the topmost
        invite in the UI).
        """
        self.return_to_homescreen()
        self.swipe_to_find_element('//android.view.View[contains(@content-desc, "You have been invited")]').click()
        self.driver.find_element(by=AppiumBy.XPATH, value='//android.widget.Button[@content-desc="ACCEPT INVITE"]').click()



import functools
import logging
from datetime import datetime

logger = logging.getLogger(__name__)

def log_action(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        module_name = func.__module__
        logger = logging.getLogger(module_name)
        logger.info(f"Action {func.__name__} initiated")
        return func(self, *args, **kwargs)
    return wrapper



import logging
from datetime import datetime
from os import makedirs
from os.path import dirname, abspath
from pathlib import Path
from sys import stdout

logger = logging.getLogger(__name__)


#####################
# LOGGING INIT CODE #
#####################

# define needed folders
PROJECT_ROOT = dirname(dirname(dirname(abspath(__file__))))
LOG_FOLDER = Path(PROJECT_ROOT) / 'logs'
makedirs(LOG_FOLDER, exist_ok=True)
CACHE_FOLDER = Path(PROJECT_ROOT) / 'cache'
makedirs(CACHE_FOLDER, exist_ok=True)


# logging helpers
def configure_default_logging():
    now = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    logging.basicConfig(
        handlers=[
            logging.FileHandler(Path(LOG_FOLDER) / f'{now}.log'),
            logging.StreamHandler(stdout)
        ],
        level=logging.INFO,
        format='%(asctime)s [%(levelname)s] [%(name)s] %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')


def log_error_and_raise_exception(logger, msg):
    logger.error(msg)
    raise Exception(msg)



import os
import subprocess
from os.path import basename, dirname, isfile
from typing import Optional
from uuid import uuid4

from puma.utils import CACHE_FOLDER, logger, log_error_and_raise_exception

FFMPEG_COMMAND = 'ffmpeg'


def _ffmpeg_is_installed() -> bool:
    completed_process = subprocess.run(f'{FFMPEG_COMMAND} -h', shell=True, capture_output=True)
    return completed_process.returncode == 0 and 'usage: ffmpeg [options]' in completed_process.stdout.decode()


def concat_videos(videos_to_concat, video_path) -> Optional[subprocess.CompletedProcess]:
    """
    A method to concat multiple video into one longer video.
    This can be useful when recording multiple actions, or when recording a video longer than the duration limit allows.

    :param videos_to_concat: The list of video files to concatenate
    :param video_path: the desired output path
    :return: the CompletedProcess for the FFMPEG command
    """
    if videos_to_concat is None or len(videos_to_concat) == 0:
        logger.error(f'Cannot concat videos: no videos provided!')
        return None
    # create folder for output file if needed
    if dirname(video_path):
        os.makedirs(dirname(video_path), exist_ok=True)
    concat_list = f'{CACHE_FOLDER}/{uuid4()}-ffmpegconcatlist.txt'
    try:
        # to concat videos ffmpeg needs an input file with all videos to concat on separate lines
        with open(concat_list, 'w') as file:
            file.write('\n'.join([f"file '{video}'" for video in videos_to_concat]))
        # ffmpeg command
        command = f'{FFMPEG_COMMAND} -f concat -safe 0 -i {concat_list} -c copy {video_path}'
        logger.info(f'Running ffmpeg with command: {command}')
        return subprocess.run(command, shell=True, capture_output=True)
    finally:
        pass
        if os.path.isfile(concat_list):
            os.remove(concat_list)


def stitch_video_horizontal(video_files: [str], output_path: str) -> subprocess.CompletedProcess:
    """
    A method to stitch videos together horizontally using ffmpeg.
    This can be useful when you want to record the screens of multiple devices, and you want to stitch multiple
    recordings together.

    Make sure ffmpeg is installed. By default the command 'ffmpeg' is used. If ffmpeg is not on your path, you need to
    set the path to the ffmpeg binary by setting the video_stitch.FFMPEG_COMMAND string.

    :param video_files: The video files to be stitched
    :param output_path: the path to the output file
    :return: the CompletedProcess for the FFMPEG command
    """
    if video_files is None or len(video_files) < 2:
        log_error_and_raise_exception(logger, f'Expected at least 2 videos to stitch together, got: {video_files}')
    for video_file in video_files:
        if not isfile(video_file):
            log_error_and_raise_exception(logger, f'given video file {video_file} does not exist')
    if not output_path:
        log_error_and_raise_exception(logger, f'Expected a valid output path, got {output_path}')

    logger.info(f'Stitching videos together horizontally: videos {[basename(f) for f in video_files]} into output '
                f'file {output_path}')
    # create folder for output file if needed
    os.makedirs(dirname(output_path), exist_ok=True)
    # check if ffmpeg is installed
    if not _ffmpeg_is_installed():
        log_error_and_raise_exception(logger, 'ffmpeg not installed, please install ffmpeg and add it to the path')
    # create ffmpeg command
    video_arguments = ' '.join([f'-i "{video_file}"' for video_file in video_files])
    command = f'{FFMPEG_COMMAND} -y {video_arguments} -filter_complex hstack=inputs={len(video_files)} "{output_path}"'
    logger.info(f'Running ffmpeg with command: {command}')
    return subprocess.run(command, shell=True, capture_output=True)



from time import sleep
from typing import Dict

from appium.webdriver.common.appiumby import AppiumBy

from puma.apps.android import log_action
from puma.apps.android.appium_actions import AndroidAppiumActions, supported_version

OPEN_CAMERA_PACKAGE = 'net.sourceforge.opencamera'


@supported_version("1.53.1")
class OpenCameraActions(AndroidAppiumActions):
    def __init__(self,
                 device_udid,
                 desired_capabilities: Dict[str, str] = None,
                 implicit_wait=1,
                 appium_server='http://localhost:4723'):
        AndroidAppiumActions.__init__(self,
                                      device_udid,
                                      OPEN_CAMERA_PACKAGE,
                                      desired_capabilities=desired_capabilities,
                                      implicit_wait=implicit_wait,
                                      appium_server=appium_server)

    def _click_shutter(self):
        """
        Presses the shutter to take a picture or start/stop recording.
        """
        xpath = '//android.widget.ImageButton[@resource-id="net.sourceforge.opencamera:id/take_photo"]'
        shutter = self.driver.find_element(by=AppiumBy.XPATH, value=xpath)
        shutter.click()

    def _switch_to_photo_mode(self):
        """
        If in video mode, switches to photo mode.
        """
        photo_mode_button = '//android.widget.ImageButton[@content-desc="Switch to photo mode"]'
        if self.is_present(photo_mode_button):
            button = self.driver.find_element(by=AppiumBy.XPATH, value=photo_mode_button)
            button.click()

    def _switch_to_video_mode(self):
        """
        If in photo mode, switches to video mode.
        """
        video_mode_button = '//android.widget.ImageButton[@content-desc="Switch to video mode"]'
        if self.is_present(video_mode_button):
            button = self.driver.find_element(by=AppiumBy.XPATH, value=video_mode_button)
            button.click()

    @log_action
    def take_picture(self):
        """
        Takes a picture.
        """
        self._switch_to_photo_mode()
        self._click_shutter()

    @log_action
    def switch_camera(self):
        """
        Switches between the front and back camera.
        """
        xpath = '//android.widget.ImageButton[@resource-id="net.sourceforge.opencamera:id/switch_camera"]'
        switch_camera_button = self.driver.find_element(by=AppiumBy.XPATH, value=xpath)
        switch_camera_button.click()

    @log_action
    def take_video(self, duration: int):
        """
        Takes a video of a given duration.
        """
        self._switch_to_video_mode()
        self._click_shutter()
        try:
            sleep(duration)
        finally:
            self._click_shutter()

    @log_action
    def zoom(self, zoom_amount: float):
        """
        Zooms to a given level between 0.0 (zoomed out completely) and 1.0 (zoomed in completely).
        """
        if zoom_amount < 0 or zoom_amount > 1:
            raise ValueError(f'Invalid zoom level: {zoom_amount}. Zoom level needs to be between 0.0 and 1.0')
        xpath = '//android.widget.SeekBar[@content-desc="Zoom"]'
        zoom_slider = self.driver.find_element(by=AppiumBy.XPATH, value=xpath)
        x = zoom_slider.location.get('x')
        y = zoom_slider.location.get('y')
        width = zoom_slider.size.get('width')
        height = zoom_slider.size.get('height')
        # We need to click slightly inside the bounds of the slider, so we don't use the full height
        height = int(0.95 * height)
        x_to_tap = x + (width // 2)
        y_to_tap = y + (height * (1 - zoom_amount))
        self.driver.tap([(x_to_tap, y_to_tap)])



import os.path

from pytesseract import pytesseract

from puma.utils import log_error_and_raise_exception, logger


class BoundingBox:
    """
    Class describing a rectangle in an image.
    properties:
        x,y: the top right corner
        width, height: the dimensions
        middle: a tuple describing the center of the rectangle
    """

    def __init__(self, x: int, y: int, width: int, height: int):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.middle = (x + (width / 2), y + (height / 2))

    def __str__(self):
        return f'BoundingBox([{self.x},{self.y}], [{self.width}x{self.height}], center: {self.middle})'

    def __repr__(self):
        return self.__str__()


class RecognizedText:
    """
    Class describing text recognized by Tesseract.
    properties:
        text: the text that was recognized
        confidence: the confidence reported by tesseract
        bounding_box: the BoundingBox of the area in which the text was recognized
    """

    def __init__(self, text: str, x: int, y: int, width: int, height: int, confidence: int):
        self.text = text
        self.bounding_box = BoundingBox(x, y, width, height)
        self.confidence = confidence

    def __str__(self):
        return f'RecognizedText(text:{self.text}, bounds:{self.bounding_box})'

    def __repr__(self):
        return self.__str__()


def recognize_text(path_to_image: str) -> [RecognizedText]:
    """
    Analyzes an image and returns a list of all text recognized by Teseract
    :param path_to_image: path to the image
    :return: a list of all RecognizedText object
    """
    if not os.path.exists(path_to_image):
        log_error_and_raise_exception(logger, f'Could not analyze image because file does not exist: {path_to_image}')
    data = pytesseract.image_to_data(path_to_image, output_type='dict')
    all_text = []
    for i in range(0, len(data['text']) - 1):
        text = RecognizedText(data['text'][i], data['left'][i], data['top'][i], data['width'][i], data['height'][i],
                              data['conf'][i])
        all_text.append(text)
    return all_text


def find_text(path_to_image: str, text_to_find: str) -> [RecognizedText]:
    """
    Filters the output of recognize_text for a given text to find (case-insensitive).
    :param path_to_image: path to the image
    :param text_to_find: text to find in image
    :return: All RecognizedText objects containing the text to find
    """
    all_text = recognize_text(path_to_image)
    return [found_text for found_text in all_text if text_to_find.lower() in found_text.text.lower()]



from typing import Dict

from appium.webdriver.common.appiumby import AppiumBy

from puma.apps.android import log_action
from puma.apps.android.appium_actions import AndroidAppiumActions, supported_version

GOOGLE_CHROME_PACKAGE = 'com.android.chrome'

@supported_version("124.0.6367.219")
class GoogleChromeActions(AndroidAppiumActions):
    def __init__(self,
                 device_udid,
                 desired_capabilities: Dict[str, str] = None,
                 impicit_wait=1,
                 appium_server='http://localhost:4723'):
        AndroidAppiumActions.__init__(self,
                                      device_udid,
                                      GOOGLE_CHROME_PACKAGE,
                                      desired_capabilities=desired_capabilities,
                                      implicit_wait=impicit_wait,
                                      appium_server=appium_server)

    @log_action
    def go_to(self, url_string: str, new_tab: bool = False):
        """
        Enters the text as stated in the url_string parameter.
        The user has the option to open a new tab first.
        :param url_string: the argument to pass to the address bar
        :param new_tab: whether to open a new tab first
        """
        search_box_xpath = '//android.widget.EditText[@resource-id="com.android.chrome:id/search_box_text"]'
        if self.is_present(search_box_xpath):
            self.driver.find_element(by=AppiumBy.XPATH, value=search_box_xpath).click()

        if new_tab:
            switch_tab_xpath = '//android.widget.ImageButton[contains(@content-desc, "Switch")]'
            new_tab_xpath = '//android.widget.TextView[@resource-id="com.android.chrome:id/new_tab_view_desc"]'
            self.driver.find_element(by=AppiumBy.XPATH, value=switch_tab_xpath).click()
            self.driver.find_element(by=AppiumBy.XPATH, value=new_tab_xpath).click()
            self.driver.find_element(by=AppiumBy.XPATH, value=search_box_xpath).click()

        url_bar_xpath = '//android.widget.EditText[@resource-id="com.android.chrome:id/url_bar"]'
        url_bar = self.driver.find_element(by=AppiumBy.XPATH, value=url_bar_xpath)
        url_bar.click()
        url_bar.send_keys(url_string)
        self.driver.press_keycode(66)

    @log_action
    def bookmark_page(self):
        """
        Bookmarks the current page.
        """
        three_dots_xpath = '//android.widget.ImageButton[@content-desc="Customize and control Google Chrome"]'
        bookmark_xpath = '//android.widget.ImageButton[lower-case(@content-desc)="bookmark"]'
        self.driver.find_element(by=AppiumBy.XPATH, value=three_dots_xpath).click()
        self.driver.find_element(by=AppiumBy.XPATH, value=bookmark_xpath).click()

    @log_action
    def load_bookmark(self):
        """
        Load the first saved bookmark in the folder 'Mobile Bookmarks'.
        """
        three_dots_xpath = '//android.widget.ImageButton[@content-desc="Customize and control Google Chrome"]'
        bookmarks_xpath = '//android.widget.TextView[@resource-id="com.android.chrome:id/menu_item_text" and @text="Bookmarks"]'
        mobile_bookmarks_xpath = '//android.widget.TextView[@resource-id="com.android.chrome:id/title" and @text="Mobile bookmarks"]'
        first_bookmark_xpath = '//android.widget.LinearLayout[@resource-id="com.android.chrome:id/container"]'
        self.driver.find_element(by=AppiumBy.XPATH, value=three_dots_xpath).click()
        self.driver.find_element(by=AppiumBy.XPATH, value=bookmarks_xpath).click()
        if self.is_present(mobile_bookmarks_xpath):
            self.driver.find_element(by=AppiumBy.XPATH, value=mobile_bookmarks_xpath).click()
        self.driver.find_element(by=AppiumBy.XPATH, value=first_bookmark_xpath).click()

    @log_action
    def switch_to_tab(self, num_tab: int = 1):
        """
        Switches to another tab, by default the first open tab.
        :param num_tab: the number of the tab to open
        """
        switch_tab_xpath = '//android.widget.ImageButton[@content-desc="Switch or close tabs"]'
        self.driver.find_element(by=AppiumBy.XPATH, value=switch_tab_xpath).click()
        self.driver.find_element(by=AppiumBy.XPATH, value=f'(//android.widget.FrameLayout[@resource-id="com.android.chrome:id/content_view"])[{num_tab}]').click()

    @log_action
    def go_to_incognito(self, url_string: str):
        """
        Opens an incognito window and enters the url_string to the address bar.
        :param url_string: the input to pass to the address bar
        """
        three_dots_xpath = '//android.widget.ImageButton[contains(@content-desc, "Customize")]'
        incognito_tab_xpath = '//android.widget.TextView[@resource-id="com.android.chrome:id/title" and @text="New Incognito tab"]'
        self.driver.find_element(by=AppiumBy.XPATH, value=three_dots_xpath).click()
        self.driver.find_element(by=AppiumBy.XPATH, value=incognito_tab_xpath).click()
        self.go_to(url_string)



import random
import threading
import time
from time import sleep
from typing import Optional, Callable

import geopy.distance
import gpxpy
import requests
from appium.webdriver.webdriver import WebDriver
from geopy import Point
from gpxpy.gpx import GPXTrackPoint

from puma.apps.android.appium_actions import AndroidAppiumActions


class RouteSimulator:
    """
    Class that allows you to spoof the location along a given route on a device available through appium.
    Starting a route is done with the execute_route_(...) methods. See their documentation on how to call these methods.
    The speed can be changed while the route is being traveled. See update_speed().
    Location spoofing can be stopped with stop_route().
    """

    def __init__(self, driver: WebDriver | AndroidAppiumActions,
                 target_speed: int,
                 absolute_speed_variance: int = None,
                 relative_speed_variance: float = None,
                 location_update_interval: float = 1):
        """
        :param driver: the Appium driver or the Puma Actions object (which contains an Appium driver).
        :param target_speed: the speed at which should be traveled when a route is started.
        :param absolute_speed_variance: (optional) the variance in speed, in kmph. see update_speed()
        :param relative_speed_variance: (optional) the relative variance in speed. see update_speed()
        :param location_update_interval: (optional) how often the location needs to be updates. default once per second
        """
        # dynamic fields
        if isinstance(driver, WebDriver):
            self.driver = driver
        else:
            self.driver = driver.driver
        self.location_update_interval = location_update_interval
        # initial speed is 0
        self.update_speed(target_speed,
                          variance_absolute=absolute_speed_variance,
                          variance_relative=relative_speed_variance)
        # fields relating to current and future locations
        self._next_point = None
        self._current_location: Optional[Point] = None
        self._next_locations: [Point] = []
        self._last_location_update: Optional[float] = None

    def update_speed(self, speed: float, variance_absolute: int = 0, variance_relative: float = 0.0):
        """
        Updates the speed at which the route is being traveled. Optionally, the speed can be variable by giving
        either an absolute or relative variance.
        This method can be called to update the speed as a route is being traveled.
        :param speed: The (average) speed the route will be traversed.
        :param variance_absolute: The absolute variance in speed. A value of 5 means the speed will vary between -5kmph
        and +5kmph of the given speed.
        :param variance_relative: The relative variance in speed. A value of 0.05 means the speed will vary between 95%
        and 105% of the given speed.
        """
        self._current_speed = speed
        if variance_absolute:
            self._target_speed_lower_bound = max(0.0, speed - variance_absolute)
            self._target_speed_upper_bound = max(0.0, speed + variance_absolute)
            return
        if variance_relative:
            self._target_speed_lower_bound = speed * max(0.0, 1 - variance_relative)
            self._target_speed_upper_bound = speed * max(0.0, 1 + variance_relative)
            return
        # no variance: constant speed
        self._target_speed_lower_bound = speed
        self._target_speed_upper_bound = speed

    def execute_route_with_points(self, point_list: [Point]):
        """
        This method executes a route by updating the GPS location of the device.
        The points it will follow are defined in the list provided by the user.
        :param point_list: the list with points of the route
        """
        self._next_locations = point_list
        self._current_location = None
        self._start_route()

    def execute_route_with_gpx(self, gpx_file: str):
        """
        This method executes a route by updating the GPS location of the device.
        The points it will follow are defined in the gpx file provided by the user.
        :param gpx_file: path to a gpx file defining the route to travel.
        """
        # Parse GPX to point list
        points: [GPXTrackPoint] = []
        with open(gpx_file, 'r') as opened_file:
            gpx = gpxpy.parse(opened_file)
            for track in gpx.tracks:
                for segment in track.segments:
                    points.extend(segment.points)
        # transform GPXTrackPoint to Point objects. Do not include p.elevation because calculating distances between
        # points with an elevation is not supported by geopy
        points = [Point(p.latitude, p.longitude, 0) for p in points]
        self.execute_route_with_points(points)

    def execute_route_with_queries(self, start_loc: str, destination: str, transport_mode: str,
                                   start_visual_function: Callable[[str, str], None] = None):
        """
        This method executes a route by updating the GPS location of the device.
        The points this route will follow are calculated based upon the two locations provided by the user.
        The locations will be queried with nominatim, the route is calculated with OSM, and the nodes
        are translated by Overpass.
        :param start_loc: the location to start from as string value
        :param destination: the destination location as string value
        :param transport_mode: the mode of transportation that is used, this can only be: 'car', 'foot', or 'bike'
        :param start_visual_function: an optional method that will be called after the route is calculated, but before the route points are being updated,
        this can be used to start the directions in a navigation based application after the start location has been set.
        """
        config = dict(user_agent="Maps")
        cls = geopy.get_geocoder_for_service("nominatim")
        geocoder = cls(**config)
        start_location = geocoder.geocode(start_loc)
        destination_location = geocoder.geocode(destination)
        start_lat = start_location.latitude
        start_lon = start_location.longitude
        end_lat = destination_location.latitude
        end_lon = destination_location.longitude
        self.driver.set_location(start_lat, start_lon)

        URL = f"http://router.project-osrm.org/route/v1/{transport_mode}/{start_lon},{start_lat};{end_lon},{end_lat}?alternatives=false&annotations=nodes"  # car / bike / foot
        r = requests.get(url=URL)
        data = r.json()
        nodes = data["routes"][0]["legs"][0]["annotation"]["nodes"]
        overpass_query = f"[out:json];({''.join('node(' + str(x) + ');' for x in nodes)});(._;>;);out;"
        r2 = requests.post(url='https://overpass-api.de/api/interpreter', verify=False, data=overpass_query)
        data2 = r2.json()

        point_dict = {}
        point_list = []
        for element in data2['elements']:
            point_dict.update({element['id']: Point(element['lat'], element['lon'])})
        for x in nodes:
            point_list.append(point_dict.get(x))

        if start_visual_function is not None:
            start_visual_function(destination, transport_mode)
        self.execute_route_with_points(point_list)

    def _start_route(self):
        """
        Starts two new threads to update the location according to the given route.
        """
        consumer_thread = threading.Thread(target=self._route_loop)
        consumer_thread.daemon = True
        consumer_thread.start()
        consumer_thread = threading.Thread(target=self._appium_loop)
        consumer_thread.daemon = True
        consumer_thread.start()

    def stop_route(self):
        """
        Stops traversing the current route. This will clear all stored locations, so the route cannot be resumed.
        If you which to pause while on the route, simply update the speed to 0.
        """
        self._next_locations = None
        self._current_location = None

    # The following methods are for internal use only, and are for traveling along the loaded route.
    def _travel_distance(self, distance_to_travel):
        """
        This method ONLY updates self._current_location along the loaded route, given a distance that should be
        traveled. This methods does nothing if there are no next locations, or if the given distance is not strictly
        positive.
        """
        # only travel if there's a waypoint as next location, and if distance is > 0
        if not self._next_locations or distance_to_travel <= 0:
            return
        distance_to_next_point = geopy.distance.geodesic(self._current_location, self._next_locations[0]).m
        # distance to next point < distance to travel: recursive function call.
        if distance_to_next_point < distance_to_travel:
            self._current_location = self._next_locations.pop(0)
            self._travel_distance(distance_to_travel - distance_to_next_point)
        else:
            # travel the required distance along the line between the current position and the next position
            percent_to_go = distance_to_travel / distance_to_next_point
            pos1 = self._current_location
            pos2 = self._next_locations[0]
            next_pos = Point(
                pos1.latitude + (pos2.latitude - pos1.latitude) * percent_to_go,
                pos1.longitude + (pos2.longitude - pos1.longitude) * percent_to_go,
                pos1.altitude + (pos2.altitude - pos1.altitude) * percent_to_go
            )
            self._current_location = next_pos

    def _update_location(self):
        """
        This method ONLY updates the _current_location and _last_location_update fields. it does not spoof the location.
        This method will look at the current speed and how long ago the location was last updated, and travel a distance
        according to said speed and duration.
        This method does nothing if there is no next location.
        """
        if not self._next_locations:
            return
        now = time.time()
        if not self._current_location:
            self._current_location = self._next_locations.pop(0)
        else:
            time_elapsed = now - self._last_location_update
            # calculate current speed with variance
            self._current_speed = random.uniform(self._target_speed_lower_bound, self._target_speed_upper_bound)
            distance_to_travel = time_elapsed * ((self._current_speed * 1000) / 3600)  # km/h to m/s
            self._travel_distance(distance_to_travel)
        self._last_location_update = now

    def _route_loop(self):
        """
        This method will travel the loaded route, and stop looping when the last point is reached.
        The loop will attempt to execute every location_update_interval seconds (with a minimum of once every 0.1s)
        """
        while self._next_locations:
            start = time.time()
            self._update_location()
            stop = time.time()
            delay = stop - start
            sleep(max(self.location_update_interval - delay, 0.1))

    def _appium_loop(self):
        """
        This method will loop and spoof the location through appium
        It will exit if there is no set _current_location().
        The loop will attempt to execute every location_update_interval seconds (with a minimum of once every 0.1s)
        """
        while self._current_location:
            start = time.time()
            self.driver.set_location(self._current_location.latitude,
                                     self._current_location.longitude,
                                     self._current_location.altitude,
                                     self._current_speed, 5)
            stop = time.time()
            delay = stop - start
            sleep(max(self.location_update_interval - delay, 0.1))



from setuptools import setup, find_packages

from puma.utils import PROJECT_ROOT
from puma.version import version

setup(
    name="pumapy",
    version=version,
    description="",
    long_description=f"{open(f'{PROJECT_ROOT}/README.md').read()}",
    long_description_content_type="text/markdown",
    author="Netherlands Forensic Institute",
    author_email="netherlandsforensicinstitute@users.noreply.github.com",
    url="https://github.com/NetherlandsForensicInstitute/puma",
    license="EUPL-1.2",
    packages=find_packages(include=['puma*']),
    test_suite="test",
    install_requires=[
        "appium-python-client~=4.3.0",
        "Pillow==10.4.0",
        "pytesseract==0.3.10",
        "geopy~=2.4.1",
        "setuptools~=80.9.0",
        "gpxpy~=1.6.2",
        "adb_pywrapper~=1.0.3",
        "requests~=2.32.3"
    ],
)



import logging
import sys

from puma.utils import configure_default_logging

MAIN_MODULE = sys.modules.get('__main__')

def _pattern_in_main(pattern):
    if hasattr(MAIN_MODULE, '__file__'):
        return MAIN_MODULE and pattern in MAIN_MODULE.__file__
    return False

def _is_running_as_main() -> bool:
    """
    Returns True if this package is being run as the main module
    via `python -m puma`.
    """
    return _pattern_in_main('puma/apps')

def _is_running_as_test() -> bool:
    return _pattern_in_main('unittest_runner')

def _is_running_in_jupyter_notebook():
    if MAIN_MODULE is not None:
        if not hasattr(MAIN_MODULE, '__file__'):
            return True
        if 'ipykernel' in str(type(MAIN_MODULE)):
            return True
    return False


configure_default_logging()

# Only configure logging when Puma is not run from another application
if (
        _is_running_as_main() or
        _is_running_as_test() or
        _is_running_in_jupyter_notebook()
):
        configure_default_logging()
else:
    logger = logging.getLogger("puma")
    logger.addHandler(logging.NullHandler())



from typing import Dict

from appium.webdriver.common.appiumby import AppiumBy

from puma.apps.android import log_action, logger
from puma.apps.android.appium_actions import AndroidAppiumActions, supported_version

GOOGLE_CAMERA_PACKAGE = 'com.google.android.GoogleCamera'


@supported_version("9.8.102")
class GoogleCameraActions(AndroidAppiumActions):
    def __init__(self,
                 device_udid,
                 desired_capabilities: Dict[str, str] = None,
                 implicit_wait=1,
                 appium_server='http://localhost:4723'):
        AndroidAppiumActions.__init__(self,
                                      device_udid,
                                      GOOGLE_CAMERA_PACKAGE,
                                      desired_capabilities=desired_capabilities,
                                      implicit_wait=implicit_wait,
                                      appium_server=appium_server)

    @log_action
    def take_picture(self):
        """
        Takes a single picture.
        """
        xpath = '//android.widget.ImageButton[@resource-id="com.google.android.GoogleCamera:id/shutter_button"]'
        shutter = self.driver.find_element(by=AppiumBy.XPATH, value=xpath)
        shutter.click()

    @log_action
    def switch_camera(self):
        """
        Switches between the front and rear camera.
        """
        xpath = '//android.widget.ImageButton[@resource-id="com.google.android.GoogleCamera:id/camera_switch_button"]'
        button = self.driver.find_element(by=AppiumBy.XPATH, value=xpath)
        button.click()
